
#include "defs.h"
#include "padarea.h"

G_DEFINE_TYPE(KpPadArea, kp_padarea, GTK_TYPE_DRAWING_AREA);

#ifndef MIN
#define MIN(a,b) (((a) < (b)) ? (a) : (b))
#endif

#ifndef MAX
#define MAX(a,b) (((a) > (b)) ? (a) : (b))
#endif

static gint stroke_add_remove_signal_id = 0;

static void kp_padarea_free_stroke(GList *stroke)
{
    GList *tmp_list = stroke;
    while(tmp_list)
    {
	g_free(tmp_list->data);
	tmp_list = tmp_list->next;
    }
    g_list_free(stroke);
}

static void kp_padarea_annotate_stroke(KpPadArea *self, GList *stroke, gint index)
{
    GdkPoint *cur, *old;

    /* Annotate the stroke with the stroke number - the algorithm
     * for placing the digit is pretty simple. The text is inscribed
     * in a circle tangent to the stroke. The circle will be above
     * and/or to the left of the line */
    if(stroke)
    {
	old = (GdkPoint*)stroke->data;

	do
	{
	    cur = (GdkPoint*)stroke->data;
	    stroke = stroke->next;
	}
	while(stroke && abs(cur->x - old->x) < 5 && abs (cur->y - old->y) < 5);

	if(stroke)
	{
	    char buffer[16];
	    PangoLayout *layout;
	    int swidth, sheight;
	    gint16 x, y;
	    double r;
	    double dx = cur->x - old->x;
	    double dy = cur->y - old->y;
	    double dl = sqrt(dx*dx+dy*dy);
	    int sign = (dy <= dx) ? 1 : -1;
	    GdkRectangle update_area;
	    GtkWidget *w = GTK_WIDGET(self);

	    sprintf(buffer, "%d", index);
	    layout = gtk_widget_create_pango_layout(GTK_WIDGET(self), buffer);
	    pango_layout_get_pixel_size(layout, &swidth, &sheight);

	    r = sqrt(swidth*swidth + sheight*sheight);

	    x = 0.5 + old->x + 0.5*r*dx/dl + sign * 0.5*r*dy/dl;
	    y = 0.5 + old->y + 0.5*r*dy/dl - sign * 0.5*r*dx/dl;

	    x -= swidth/2;
	    y -= sheight/2;

	    update_area.x = x;
	    update_area.y = y;
	    update_area.width = swidth;
	    update_area.height = sheight;

	    x = CLAMP(x, 0, w->allocation.width - swidth);
	    y = CLAMP(y, 0, w->allocation.height - sheight);

	    cairo_t *cr = gdk_cairo_create(GTK_WIDGET(self)->window);
	    cairo_move_to(cr, x, y);
	    pango_cairo_show_layout(cr, layout);
	    cairo_stroke(cr);
	    cairo_destroy(cr);
	}
    }
}

static void kp_padarea_draw_line(KpPadArea *self, GdkPoint *p0, GdkPoint *p1)
{
    cairo_t *cr = gdk_cairo_create(GTK_WIDGET(self)->window);
    cairo_set_source_rgb(cr, 0.0, 0.0, 0.0);
    cairo_move_to(cr, p0->x, p0->y);
    cairo_line_to(cr, p1->x, p1->y);
    cairo_stroke(cr);
    cairo_destroy(cr);
}

static void kp_padarea_draw_stroke(KpPadArea *self, GList *s)
{
    for(;s != g_list_last(s); s = g_list_next(s))
    {
	GdkPoint *p0 = (GdkPoint *)s->data;
	GdkPoint *p1 = (GdkPoint *)g_list_next(s)->data;
	kp_padarea_draw_line(self, p0, p1);
    }
}

static void kp_padarea_reset(KpPadArea *self)
{
    GList *l = self->strokes;
    int index = 1;
    GtkWidget *widget = GTK_WIDGET(self);

    guint16 width = widget->allocation.width;
    guint16 height = widget->allocation.height;

    cairo_t *cr = gdk_cairo_create(GTK_WIDGET(self)->window);
    cairo_set_source_rgb(cr, 1.0, 1.0, 1.0);
    cairo_rectangle(cr, 0, 0, width, height);
    cairo_fill(cr);
    cairo_destroy(cr);

    for(l = self->strokes; l; l = g_list_next(l))
    {
	if(self->annotate)
	    kp_padarea_annotate_stroke(self, l->data, index++);

	kp_padarea_draw_stroke(self, l->data);
    }
}

static int kp_padarea_configure_cb(GtkWidget *w, GdkEventConfigure *event, gpointer data)
{
    KpPadArea *self = KP_PADAREA(w);

    kp_padarea_reset(self);

    return TRUE;
}

static int kp_padarea_expose_cb(GtkWidget *w, GdkEventExpose *event, gpointer data)
{
    KpPadArea *self = KP_PADAREA(w);
    kp_padarea_reset(self);
    return TRUE;
}

static int kp_padarea_button_press_cb(GtkWidget *w, GdkEventButton *event, gpointer data)
{
    KpPadArea *self = KP_PADAREA(w);
    if(event->button == 1)
    {
	GdkPoint *p = g_new(GdkPoint, 1);
	p->x = event->x;
	p->y = event->y;
	self->curstroke = g_list_append(self->curstroke, p);
	self->instroke = TRUE;
	kp_padarea_draw_line(self, p, p);
    }
    return TRUE;
}

static int kp_padarea_button_release_cb(GtkWidget *w, GdkEventButton *event, gpointer data)
{
    KpPadArea *self = KP_PADAREA(w);
    if(self->annotate)
	kp_padarea_annotate_stroke(self, self->curstroke, g_list_length (self->strokes) + 1);

    self->strokes = g_list_append(self->strokes, self->curstroke);
    self->curstroke = NULL;
    self->instroke = FALSE;
    g_signal_emit(w, stroke_add_remove_signal_id, 0);
    return TRUE;
}

static int kp_padarea_motion_cb(GtkWidget *w, GdkEventMotion *event, gpointer data)
{
    gint x, y;
    GdkModifierType state;
    KpPadArea *self = KP_PADAREA(w);

    if(event->is_hint)
    {
	gdk_window_get_pointer(w->window, &x, &y, &state);
    }
    else
    {
	x = event->x;
	y = event->y;
	state = event->state;
    }

    if(self->instroke && state & GDK_BUTTON1_MASK)
    {
	GdkPoint *p0 = (GdkPoint *)g_list_last(self->curstroke)->data;
	GdkPoint *pt = g_new(GdkPoint, 1);
	pt->x = x;
	pt->y = y;

	kp_padarea_draw_line(self, p0, pt);
	self->curstroke = g_list_append(self->curstroke, pt);
    }

    return TRUE;
}

static void kp_padarea_class_init(KpPadAreaClass *klass)
{
    stroke_add_remove_signal_id = g_signal_new("stroke_add_remove",
		G_TYPE_OBJECT, G_SIGNAL_RUN_LAST,
		0, NULL, NULL,
		g_cclosure_marshal_VOID__VOID,
		G_TYPE_NONE, 0);
}

static void kp_padarea_init(KpPadArea *self)
{
    GtkWidget *w = GTK_WIDGET(self);

    g_signal_connect(self, "configure_event", G_CALLBACK(kp_padarea_configure_cb), NULL);
    g_signal_connect(self, "expose_event", G_CALLBACK(kp_padarea_expose_cb), NULL);
    g_signal_connect(self, "button_press_event", G_CALLBACK(kp_padarea_button_press_cb), NULL);
    g_signal_connect(self, "button_release_event", G_CALLBACK(kp_padarea_button_release_cb), NULL);
    g_signal_connect(self, "motion_notify_event", G_CALLBACK(kp_padarea_motion_cb), NULL);

    gtk_widget_set_events(w, GDK_EXPOSURE_MASK | GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK | GDK_POINTER_MOTION_MASK | GDK_POINTER_MOTION_HINT_MASK);

    self->strokes = NULL;
    self->curstroke = NULL;
    self->instroke = FALSE;
    self->annotate = FALSE;
    self->line_width = 3;
}

void kp_padarea_clear(KpPadArea *self)
{
    GList *tmp_list;

    tmp_list = self->strokes;
    while(tmp_list)
    {
	kp_padarea_free_stroke(tmp_list->data);
	tmp_list = tmp_list->next;
    }
    g_list_free(self->strokes);
    self->strokes = NULL;

    g_list_free(self->curstroke);
    self->curstroke = NULL;

    kp_padarea_reset(self);
    g_signal_emit(self, stroke_add_remove_signal_id, 0);
}

void kp_padarea_undo(KpPadArea *self)
{
    if(!self->strokes)
	return;

    GList *tmp = g_list_last(self->strokes);
    self->strokes = g_list_remove_link(self->strokes, tmp);
    kp_padarea_free_stroke(tmp->data);
    g_list_free(tmp);
    kp_padarea_reset(self);
    g_signal_emit(self, stroke_add_remove_signal_id, 0);
}

void kp_padarea_annotate(KpPadArea *self, gboolean annotate)
{
    if(self->annotate != annotate)
    {
	self->annotate = annotate;
	kp_padarea_reset(self);
    }
}

